# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from enum import IntEnum
from typing import Dict

from .app_command import AppCommandInteractionDataOption
from .channel import Channel
from .guild_member import GuildMember
from .interaction_base import InteractionType
from .user_message import UserMessage
from .role import Role
from .select_menu import SelectOption
from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake, convert


class InteractionFlags(IntEnum):
    """
    :param EPHEMERAL:
        only the user receiving the message can see it
    """
    EPHEMERAL = 1 << 6


@dataclass
class ResolvedData(APIObject):
    """
    Represents a Discord Resolved Data structure

    :param users:
        Map of Snowflakes to user objects

    :param members:
        Map of Snowflakes to partial member objects

    :param roles:
        Map of Snowflakes to role objects

    :param channels:
        Map of Snowflakes to partial channel objects

    :param messages:
        Map of Snowflakes to partial message objects
    """
    users: APINullable[Dict[Snowflake, User]] = MISSING
    members: APINullable[Dict[Snowflake, GuildMember]] = MISSING
    roles: APINullable[Dict[Snowflake, Role]] = MISSING
    channels: APINullable[Dict[Snowflake, Channel]] = MISSING
    messages: APINullable[Dict[Snowflake, UserMessage]] = MISSING


@dataclass
class InteractionData(APIObject):
    """
    Represents a Discord Interaction Data structure

    :param id:
        the `ID` of the invoked command

    :param name:
        the `name` of the invoked command

    :param type:
        the `type` of the invoked command

    :param resolved:
        converted users + roles + channels

    :param options:
        the params + values from the user

    :param custom_id:
        the `custom_id` of the component

    :param component_type:
        the type of the component

    :param values:
        the values the user selected

    :param target_id:
        id of the user or message targeted by a user or message command
    """
    id: Snowflake
    name: str
    type: int

    resolved: APINullable[ResolvedData] = MISSING
    options: APINullable[AppCommandInteractionDataOption] = MISSING
    custom_id: APINullable[str] = MISSING
    component_type: APINullable[int] = MISSING
    values: APINullable[SelectOption] = MISSING
    target_id: APINullable[Snowflake] = MISSING

    def __post_init__(self):
        self.id = convert(self.id, Snowflake.from_string)
        self.resolved = convert(
            self.resolved,
            ResolvedData.from_dict,
            ResolvedData
        )
        self.options = convert(
            self.options,
            AppCommandInteractionDataOption.from_dict,
            AppCommandInteractionDataOption
        )
        self.values = convert(self.values, SelectOption.from_dict, SelectOption)
        self.target_id = convert(self.target_id, Snowflake.from_string)


@dataclass
class Interaction(APIObject):
    """
    Represents a Discord Interaction object

    :param id:
        id of the interaction

    :param application_id:
        id of the application this interaction is for

    :param type:
        the type of interaction

    :param data:
        the command data payload

    :param guild_id:
        the guild it was sent from

    :param channel_id:
        the channel it was sent from

    :param member:
        guild member data for the invoking user, including permissions

    :param user:
        user object for the invoking user, if invoked in a DM

    :param token:
        a continuation token for responding to the interaction

    :param version:
        read-only property, always `1`

    :param message:
        for components, the message they were attached to
    """
    id: Snowflake
    application_id: Snowflake
    type: InteractionType
    token: str

    version: int = 1
    data: APINullable[InteractionData] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    channel_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING
    user: APINullable[User] = MISSING
    message: APINullable[UserMessage] = MISSING

    def __post_init__(self):
        self.id = convert(self.id, Snowflake.from_string)
        self.application_id = convert(
            self.application_id, Snowflake.from_string
        )
        self.type = convert(self.type, InteractionType)
        self.data = convert(
            self.data,
            InteractionData.from_dict,
            InteractionData
        )
        self.guild_id = convert(self.guild_id, Snowflake.from_string)
        self.channel_id = convert(self.channel_id, Snowflake.from_string)
        self.member = convert(self.member, GuildMember.from_dict, GuildMember)
        self.user = convert(self.user, User.from_dict, User)
        self.message = convert(self.message, UserMessage.from_dict, UserMessage)



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from dataclasses import dataclass, MISSING
from typing import List

from pincer.objects import Integration
from pincer.objects.user import VisibilityType
from pincer.utils import APIObject, APINullable


@dataclass
class Connection(APIObject):
    """
    The connection object that the user has attached.

    :param id:
        id of the connection account

    :param name:
        the username of the connection account

    :param type:
        the service of the connection (twitch, youtube)

    :param revoked:
        whether the connection is revoked

    :param integrations:
        an array of partial server integrations

    :param verified:
        whether the connection is verified

    :param friend_sync:
        whether friend sync is enabled for this connection

    :param show_activity:
        whether activities related to this connection
        will be shown in presence updates
    """
    id: str
    name: str
    type: str
    verified: bool
    friend_sync: bool
    show_activity: bool
    visibility: VisibilityType

    revoked: APINullable[bool] = MISSING
    integrations: APINullable[List[Integration]] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""sent when a message is created in a subscribed text channel"""

from pincer.core.dispatch import GatewayDispatch
from pincer.objects import UserMessage


async def message_create_middleware(payload: GatewayDispatch):
    return "on_message", [UserMessage.from_dict(payload.data)]


def export():
    return message_create_middleware



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, Optional, List, TYPE_CHECKING

from .._config import GatewayConfig
from .guild_member import GuildMember
from .thread import ThreadMetadata
from .user import User
from .overwrite import Overwrite
from ..utils import APIObject, APINullable, MISSING, Snowflake, Timestamp

if TYPE_CHECKING:
    from pincer import Client
    from ..core.http import HTTPClient


class ChannelType(IntEnum):
    """Represents a channel its type."""
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_NEWS = 5
    GUILD_STORE = 6

    if GatewayConfig.version >= 9:
        GUILD_NEWS_THREAD = 10
        GUILD_PUBLIC_THREAD = 11
        GUILD_PRIVATE_THREAD = 12

    GUILD_STAGE_VOICE = 13


@dataclass
class Channel(APIObject):
    """
    Represents a Discord Channel Mention object

    :param id:
        the id of this channel

    :param type:
        the type of channel

    :param application_id:
        application id of the group DM creator if it is bot-created

    :param bitrate:
        the bitrate (in bits) of the voice channel

    :param default_auto_archive_duration:
        default duration for newly created threads, in minutes, to
        automatically archive the thread after recent activity, can be set to:
        60, 1440, 4320, 10080

    :param guild_id:
        the id of the guild (may be missing for some channel objects received
        over gateway guild dispatches)

    :param icon:
        icon hash
    :param last_message_id:
        the id of the last message sent in this channel (may not point to an
        existing or valid message)

    :param last_pin_timestamp:
        when the last pinned message was pinned. This may be null in events
        such as GUILD_CREATE when a message is not pinned.

    :param member:
        thread member object for the current user, if they have joined the
        thread, only included on certain API endpoints

    :param member_count:
        an approximate count of users in a thread, stops counting at 50

    :param message_count:
        an approximate count of messages in a thread, stops counting at 50

    :param name:
        the name of the channel (1-100 characters)

    :param nsfw:
        whether the channel is nsfw

    :param owner_id:
        id of the creator of the group DM or thread

    :param parent_id:
        for guild channels: id of the parent category for a channel (each
        parent category can contain up to 50 channels), for threads: id of the
        text channel this thread was created

    :param permissions:
        computed permissions for the invoking user in the channel, including
        overwrites, only included when part of the resolved data received on a
        slash command interaction

    :param permission_overwrites:
        explicit permission overwrites for members and roles

    :param position:
        sorting position of the channel

    :param rate_limit_per_user:
        amount of seconds a user has to wait before sending another message
        (0-21600); bots, as well as users with the permission manage_messages
        or manage_channel, are unaffected

    :param recipients:
        the recipients of the DM

    :param rtc_region:
        voice region id for the voice channel, automatic when set to null

    :param thread_metadata:
        thread-specific fields not needed by other channels

    :param topic:
        the channel topic (0-1024 characters)

    :param user_limit:
        the user limit of the voice channel

    :param video_quality_mode:
        the camera video quality mode of the voice channel, 1 when not present
    """

    _client: Client
    _http: HTTPClient

    id: Snowflake
    type: ChannelType

    application_id: APINullable[Snowflake] = MISSING
    bitrate: APINullable[int] = MISSING
    default_auto_archive_duration: APINullable[int] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    icon: APINullable[Optional[str]] = MISSING

    last_message_id: APINullable[Optional[Snowflake]] = MISSING
    last_pin_timestamp: APINullable[Optional[Timestamp]] = MISSING
    member: APINullable[GuildMember] = MISSING

    member_count: APINullable[int] = MISSING

    message_count: APINullable[int] = MISSING

    name: APINullable[str] = MISSING
    nsfw: APINullable[bool] = MISSING
    owner_id: APINullable[Snowflake] = MISSING
    parent_id: APINullable[Optional[Snowflake]] = MISSING
    permissions: APINullable[str] = MISSING
    permission_overwrites: APINullable[List[Overwrite]] = MISSING
    position: APINullable[int] = MISSING
    rate_limit_per_user: APINullable[int] = MISSING
    recipients: APINullable[List[User]] = MISSING
    rtc_region: APINullable[Optional[str]] = MISSING
    thread_metadata: APINullable[ThreadMetadata] = MISSING
    topic: APINullable[Optional[str]] = MISSING
    user_limit: APINullable[int] = MISSING
    video_quality_mode: APINullable[int] = MISSING

    @classmethod
    async def from_id(cls, client: Client, id: int) -> Channel:
        data = (await client.http.get(f"/guilds/{id}")) or {}
        data.update(
            {
                "_client": client,
                "_http": client.http,
                "type": ChannelType(data.pop("type"))
            }
        )

        channel_cls = _channel_type_map.get(data["type"], Channel)
        return channel_cls.from_dict(data)


def __str__(self):
    """return the discord tag when object gets used as a string."""
    return self.name or str(self.id)


@dataclass
class TextChannel(Channel):
    pass


@dataclass
class VoiceChannel(Channel):
    pass


@dataclass
class CategoryChannel(Channel):
    pass


@dataclass
class NewsChannel(Channel):
    pass


@dataclass
class ChannelMention(APIObject):
    """
    Represents a Discord Channel Mention object

    :param id:
        id of the channel

    :param guild_id:
        id of the guild containing the channel

    :param type:
        the type of channel

    :param name:
        the name of the channel
    """
    id: Snowflake
    guild_id: Snowflake
    type: ChannelType
    name: str


_channel_type_map: Dict[ChannelType, Channel] = {
    ChannelType.GUILD_TEXT: TextChannel,
    ChannelType.GUILD_VOICE: VoiceChannel,
    ChannelType.GUILD_CATEGORY: CategoryChannel,
    ChannelType.GUILD_NEWS: NewsChannel
}



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from typing import Any, List, Optional

from .presence import PresenceUpdateEvent
from pincer.objects.guild_member import GuildMember
from pincer.objects.emoji import Emoji
from pincer.objects.role import Role
from pincer.objects.sticker import Sticker
from pincer.objects.user import User
from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.timestamp import Timestamp
from pincer.utils.types import MISSING, APINullable


@dataclass
class GuildBanAddEvent(APIObject):
    """
    Sent when a user is banned from a guild.

    :param guild_id:
        id of the guild

    :param user:
        the banned user
    """
    guild_id: Snowflake
    user: User


@dataclass
class GuildBanRemoveEvent(APIObject):
    """
    Sent when a user is unbanned from a guild.

    :param guild_id:
        id of the guild

    :param user:
        the unbanned user
    """
    guild_id: Snowflake
    user: User


@dataclass
class GuildEmojisUpdateEvent(APIObject):
    """
    Sent when a guild's emojis have been updated.

    :param guild_id:
        id of the guild

    :param emojis:
        array of emojis
    """
    guild_id: Snowflake
    emojis: List[Emoji]


@dataclass
class GuildStickersUpdateEvent(APIObject):
    """
    Sent when a guild's stickers have been updated.

    :param guild_id:
        id of the guild

    :param stickers:
        array of stickers
    """
    guild_id: Snowflake
    stickers: List[Sticker]


@dataclass
class GuildIntegrationsUpdateEvent(APIObject):
    """
    Sent when a guild integration is updated.

    :param guild_id:
        id of the guild whose integrations were updated
    """
    guild_id: Snowflake


@dataclass
class GuildMemberRemoveEvent(APIObject):
    """
    Sent when a user is removed from a guild (leave/kick/ban).

    :param guild_id:
        the id of the guild

    :param user:
        the user who was removed
    """
    guild_id: Snowflake
    user: User


@dataclass
class GuildMemberUpdateEvent(APIObject):
    """
    Sent when a guild member is updated.
    This will also fire when the user object
    of a guild member changes.

    :param guild_id:
        the id of the guild

    :param roles:
        user role ids

    :param user:
        the user

    :param nick:
        nickname of the user in the guild

    :param joined_at:
        when the user joined the guild

    :param premium_since:
        when the user started boosting the guild

    :param deaf:
        whether the user is deafened in voice channels

    :param mute:
        whether the user is muted in voice channels

    :param pending:
        whether the user has not yet passed the guild's
        Membership Screening requirements
    """
    guild_id: Snowflake
    roles: List[Snowflake]
    user: User
    nick: APINullable[Optional[str]] = MISSING
    joined_at: Optional[Timestamp] = None
    premium_since: APINullable[Optional[Timestamp]] = MISSING
    deaf: APINullable[bool] = MISSING
    mute: APINullable[bool] = MISSING
    pending: APINullable[bool] = MISSING


@dataclass
class GuildMembersChunkEvent(APIObject):
    """
    Sent in response to Guild Request Members.
    You can use the `chunk_index` and `chunk_count`
    to calculate how many chunks are left for your request.

    :param guild_id:
        the id of the guild

    :param members:
        set of guild members

    :param chunk_index:
        the chunk index in the expected chunks for this response
        (0 <= chunk_index < chunk_count)

    :param chunk_count:
        the total number of expected chunks for this response

    :param not_found:
        if passing an invalid id to `REQUEST_GUILD_MEMBERS`,
        it will be returned here

    :param presences:
        if passing true to `REQUEST_GUILD_MEMBERS`, presences
        of the returned members will be here

    :param nonce:"
        the nonce used in the Guild Members Request
    """
    guild_id: Snowflake
    members: List[GuildMember]
    chunk_index: int
    chunk_count: int

    not_found: APINullable[List[Any]] = MISSING
    presences: APINullable[PresenceUpdateEvent] = MISSING
    nonce: APINullable[str] = MISSING


@dataclass
class GuildRoleCreateEvent(APIObject):
    """
    Sent when a guild role is created.

    :param guild_id:
        the id of the guild

    :param role:
        the role created
    """
    guild_id: Snowflake
    role: Role


@dataclass
class GuildRoleUpdateEvent(APIObject):
    """
    Sent when a guild role is updated.

    :param guild_id:
        the id of the guild

    :param role:
        the role updated
    """
    guild_id: Snowflake
    role: Role


@dataclass
class GuildRoleDeleteEvent(APIObject):
    """
    Sent when a guild role is deleted.

    :param guild_id:
        id of the guild

    :param role_id:
        id of the role
    """
    guild_id: Snowflake
    role_id: Snowflake



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum

from ..objects.emoji import Emoji
from ..utils import APIObject, APINullable, MISSING


class ButtonStyle(IntEnum):
    """
    Buttons come in a variety of styles to convey different types of actions.
    These styles also define what fields are valid for a button.

    Primary:
        - color: blurple
        - required_field: custom_id
    Secondary:
        - color: gray
        - required_field: custom_id
    Success:
        - color: green
        - required_field: custom_id
    Danger:
        - color: red
        - required_field: custom_id
    Link:
        - color: gray, navigates to a URL
        - required_field: url
    """
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


@dataclass
class Button(APIObject):
    """
    Represents a Discord Button object.
    Buttons are interactive components that render on messages.

    They can be clicked by users,
    and send an interaction to your app when clicked.

    :param type:
        `2` for a button

    :param style:
        one of button styles

    :param label:
        text that appears on the button, max 80 characters

    :param emoji:
        `name`, `id`, and `animated`

    :param custom_id:
        a developer-defined identifier for the button,
        max 100 characters

    :param url:
        a url for link-style buttons

    :param disabled:
        whether the button is disabled (default `False`)
    """
    type: int
    style: ButtonStyle

    label: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    custom_id: APINullable[str] = MISSING
    url: APINullable[str] = MISSING
    disabled: APINullable[bool] = False



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, List

from .application import Application
from .channel import Channel
from .guild import Guild
from .guild_member import GuildMember
from .user import User
from ..utils import APIObject, APINullable, MISSING, Timestamp


class InviteTargetType(IntEnum):
    """
    Represents the type of the invite.

    :param STREAM:
        A normal Discord invite, eg for a channel or guild.

    :param EMBEDDED_APPLICATION:
        An embedded application invite, eg poker-night etc
    """
    STREAM = 1
    EMBEDDED_APPLICATION = 2


@dataclass
class InviteStageInstance(APIObject):
    """
    Represents an invite for a Discord stages channel.

    :param members:
        the members speaking in the Stage

    :param participant_count:
        the number of users in the Stage

    :param speaker_count:
        the number of users speaking in the Stage

    :param topic:
        the topic of the Stage instance (1-120 characters)
    """

    members: List[GuildMember]
    participant_count: int
    speaker_count: int
    topic: str


@dataclass
class InviteMetadata(APIObject):
    """
    Extra information about an invite, will extend the invite object.

    :param uses:
        number of times this invite has been used

    :param max_uses:
        max number of times this invite can be used

    :param max_age:
        duration (in seconds) after which the invite expires

    :param temporary:
        whether this invite only grants temporary membership

    :param created_at:
        when this invite was created
    """
    uses: int
    max_uses: int
    max_age: int
    temporary: bool
    created_at: Timestamp


@dataclass
class Invite(APIObject):
    """
    Represents a Discord invite.

    :param channel:
        the channel this invite is for

    :param code:
        the invite code (unique ID)

    :param approximate_member_count:
        approximate count of total members, returned from the GET
        /invites/<code> endpoint when with_counts is true

    :param approximate_presence_count:
        approximate count of online members, returned from the GET
        /invites/<code> endpoint when with_counts is true

    :param expires_at:
        the expiration date of this invite, returned from the GET
        /invites/<code> endpoint when with_expiration is true

    :param inviter:
        the user who created the invite

    :param guild:
        the guild this invite is for

    :param stage_instance:
        stage instance data if there is a public Stage instance in the Stage
        channel this invite is for

    :param target_type:
        the type of target for this voice channel invite

    :param target_user:
        the user whose stream to display for this voice channel stream invite

    :param target_application:
        the embedded application to open for this voice channel embedded
        application invite
    """

    channel: Channel
    code: str

    approximate_member_count: APINullable[int] = MISSING
    approximate_presence_count: APINullable[int] = MISSING
    expires_at: APINullable[Optional[Timestamp]] = MISSING
    inviter: APINullable[User] = MISSING
    guild: APINullable[Guild] = MISSING
    stage_instance: APINullable[InviteStageInstance] = MISSING
    target_type: APINullable[InviteTargetType] = MISSING
    target_user: APINullable[User] = MISSING
    target_application: APINullable[Application] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass


@dataclass
class GatewayConfig:
    """
    This file is to make maintaining the library its gateway
        configuration easier.
    """
    socket_base_url: str = "wss://gateway.discord.gg/"
    version: int = 9
    encoding: str = "json"
    compression: str = "zlib-stream"

    @staticmethod
    def uri() -> str:
        """
        :return uri:
            The GatewayConfig's uri.
        """
        return (
            f"{GatewayConfig.socket_base_url}"
            f"?v={GatewayConfig.version}"
            f"&encoding={GatewayConfig.encoding}"
        )


events = [
    "ready", "channel_create", "channel_update", "channel_delete",
    "channel_pin_update", "thread_create", "thread_update",
    "thread_delete", "thread_member_update", "thread_members_update",
    "guild_create", "guild_update", "guild_delete", "guild_ban_add",
    "guild_ban_remove", "guild_emoji_update", "guild_stickers_update",
    "guild_integrations_update", "guild_member_add",
    "guild_member_remove", "guild_members_chunk", "guild_role_create",
    "guild_role_update", "guild_role_delete", "integration_create",
    "integration_update", "integration_delete", "invite_create",
    "messages_create", "message_update", "message_delete",
    "message_delete_bulk", "message_reaction_add",
    "message_reaction_remove", "message_reaction_remove_all",
    "message_reaction_remove_emoji", "presence_update", "stage_instance_create",
    "stage_instance_update", "stage_instance_delete", "typing_start",
    "voice_state_update", "voice_server_update", "webhooks_update"
]



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from .channel import Channel
from .guild import Guild
from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake


class WebhookType(IntEnum):
    """
    Represents the type of a webhook.

    :param INCOMING:
        Incoming Webhooks can post messages to channels with a
        generated token.

    :param CHANNEL_FOLLOWER:
        Channel Follower Webhooks are internal webhooks used with
        Channel Following to post new messages into channels.

    :param APPLICATION:
        Application webhooks are webhooks used with Interactions
    """
    INCOMING = 1
    CHANNEL_FOLLOWER = 2
    APPLICATION = 3


@dataclass
class Webhook(APIObject):
    """
    Represents a Discord channel webhook.

    :param id:
        the id of the webhook

    :param type:
        the type of the webhook

    :param channel_id:
        the channel id this webhook is for, if any

    :param name:
        the default name of the webhook

    :param avatar:
        the default user avatar hash of the webhook

    :param application_id:
        the bot/OAuth2 application that created this webhook

    :param user:
        the user this webhook was created by
        (not returned when getting a webhook with its token)

    :param token:
        the secure token of the webhook
        (returned for Incoming Webhooks)

    :param source_guild:
        the guild of the channel that this webhook is following
        (returned for Channel Follower Webhooks)

    :param source_channel:
        the channel that this webhook is following
        (returned for Channel Follower Webhooks)

    :param url:
        the url used for executing the webhook
        (returned by the webhooks OAuth2 flow)

    :param guild_id:
        the guild id this webhook is for, if any
    """

    id: Snowflake
    type: WebhookType

    channel_id: Optional[Snowflake] = None
    name: Optional[str] = None
    avatar: Optional[str] = None
    application_id: Optional[Snowflake] = None

    user: APINullable[User] = MISSING
    token: APINullable[str] = MISSING
    source_guild: APINullable[Guild] = MISSING
    source_channel: APINullable[Channel] = MISSING
    url: APINullable[str] = MISSING

    guild_id: APINullable[Optional[Snowflake]] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import List, Optional

from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake


class StickerType(IntEnum):
    """
    Displays from where the sticker comes from.

    :param STANDARD:
        Sticker is included in the default Discord sticker pack.

    :param GUILD:
        Sticker is a custom sticker from a discord server.
    """
    STANDARD = 1
    GUILD = 2


class StickerFormatType(IntEnum):
    """
    The type of the sticker.

    :param PNG:
        Sticker is of PNG format.

    :param APNG:
        Sticker is animated with APNG format.

    :param LOTTIE:
        Sticker is animated with with LOTTIE format. (vector based)
    """
    PNG = 1
    APNG = 2
    LOTTIE = 3


@dataclass
class Sticker(APIObject):
    """
    Represents a Discord sticker.

    :param description:
        description of the sticker

    :param format_type:
        type of sticker format

    :param id:
        id of the sticker

    :param name:
        name of the sticker

    :param tags:
        for guild stickers, the Discord name of a unicode emoji
        representing the sticker's expression. For standard stickers,
        a comma-separated list of related expressions.

    :param type:
        type of sticker

    :param available:
        whether this guild sticker can be used,
        may be false due to loss of Server Boosts

    :param guild_id:
        id of the guild that owns this sticker

    :param pack_id:
        for standard stickers, id of the pack the sticker is from

    :param sort_value:
        the standard sticker's sort order within its pack

    :param user:
        the user that uploaded the guild sticker
    """

    description: Optional[str]
    format_type: StickerFormatType
    id: Snowflake
    name: str
    tags: str
    type: StickerType

    available: APINullable[bool] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    pack_id: APINullable[Snowflake] = MISSING
    sort_value: APINullable[int] = MISSING
    user: APINullable[User] = MISSING


@dataclass
class StickerItem(APIObject):
    """
    Represents the smallest amount of data required to render a sticker.
    A partial sticker object.

    :param id:
        id of the sticker

    :param name:
        name of the sticker

    :param format_type:
        type of sticker format
    """

    id: Snowflake
    name: str
    format_type: StickerFormatType


@dataclass
class StickerPack(APIObject):
    """
    Represents a pack of standard stickers.

    :param id:
        id of the sticker pack

    :param stickers:
        the stickers in the pack

    :param name:
        name of the sticker pack

    :param sku_id:
        id of the pack's SKU

    :param description:
        description of the sticker pack

    :param cover_sticker_id:
        id of a sticker in the pack which is shown as the pack's icon

    :param banner_asset_id:
        id of the sticker pack's banner image
    """

    id: Snowflake
    stickers: List[Sticker]
    name: str
    sku_id: Snowflake
    description: str

    cover_sticker_id: APINullable[Snowflake] = MISSING
    banner_asset_id: APINullable[Snowflake] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass

from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.timestamp import Timestamp
from pincer.utils.types import MISSING, APINullable


@dataclass
class ChannelPinsUpdateEvent(APIObject):
    """
    Sent when a message is pinned or unpinned in a text channel.
    This is not sent when a pinned message is deleted.

    :param guild_id:
        the id of the guild

    :param channel_id:
        the id of the channel

    :param last_pin_timestamp:
        the time at which the most recent pinned message was pinned
    """
    channel_id: Snowflake

    guild_id: APINullable[Snowflake] = MISSING
    last_pin_timestamp: APINullable[Timestamp] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from typing import Optional


class PincerError(Exception):
    """Base exception class for all Pincer errors."""


class UnhandledException(PincerError):
    """
    Exception which gets thrown if an exception wasn't handled.

    Please create an issue on our github
    if this exception gets thrown.
    """

    def __init__(self, specific: str):
        super(UnhandledException, self).__init__(
            specific + " Please report this to the library devs."
        )


class NoExportMethod(PincerError):
    """
    Exception which gets raised when an `export` method is expected but not
    found in a module.
    """


class CommandError(PincerError):
    """
    Base class for exceptions which are related to commands.
    """


class CommandIsNotCoroutine(CommandError):
    """
    Exception raised when the provided command call is not a coroutine.
    """


class CommandAlreadyRegistered(CommandError):
    """
    The command which you are trying to register is already registered.
    """


class CommandDescriptionTooLong(CommandError):
    """
    The provided command description is too long, as it exceeds 100 characters.
    """


class TooManyArguments(CommandError):
    """
    A command can have a maximum of 25 arguments.
    If this number of arguments gets exceeded, this exception will be raised.
    """


class InvalidArgumentAnnotation(CommandError):
    """
    The provided argument annotation is not known, so it cannot be used.
    """


class CommandReturnIsEmpty(CommandError):
    """
    Cannot return an empty string to an interaction.
    """


class InvalidCommandGuild(CommandError):
    """
    The provided guild id not not valid.
    """


class InvalidEventName(PincerError):
    """
    Exception raised when the event name is not a valid event.
    This can be because the event name did not begin with an ``on_`` or
    because its not a valid event in the library.
    """


class InvalidUrlError(PincerError, ValueError):
    """
    Exception raised when an invalid url has been provided.
    """


class EmbedFieldError(PincerError, ValueError):
    """Exception that is raised when an embed field is too large"""

    @classmethod
    def from_desc(cls, _type: str, max_size: int, cur_size: int):
        """
        Create an instance by description.

        :param _type:
            The type/name of the field.

        :param max_size:
            The maximum size of the field.

        :param cur_size:
            The current size of the field.
        """
        return cls(
            f"{_type} can have a maximum length of {max_size}."
            f" (Current size: {cur_size})"
        )


class DispatchError(PincerError):
    """
    Base exception class for all errors which are specifically related
    to the dispatcher.
    """


class _InternalPerformReconnectError(DispatchError):
    """Internal helper exception which on raise lets the client reconnect."""


class DisallowedIntentsError(DispatchError):
    """
    Invalid gateway intent got provided.
    Make sure your client has the enabled intent.
    """


class InvalidTokenError(DispatchError, ValueError):
    """
    Exception raised when the authorization token is invalid.
    """

    def __init__(self, hint: Optional[str] = None):
        """
        :param hint:
            Additional information about the exception cause.
        """
        hint = hint or ''

        super(InvalidTokenError, self).__init__(
            "The given token is not a valid token.\n" + hint
        )


class HeartbeatError(DispatchError):
    """Exception raised due to a problem with websocket heartbeat."""


class UnavailableGuildError(PincerError):
    """
    Exception raised due to a guild being unavailable.
    This is caused by a discord outage.
    """


# Discord HTTP Errors
# `developers/docs/topics/opcodes-and-status-codes#http`


class HTTPError(PincerError):
    """HTTP Exception base class."""


class NotModifiedError(HTTPError):
    """Error code 304."""


class BadRequestError(HTTPError):
    """Error code 400."""


class UnauthorizedError(HTTPError):
    """Error code 401."""


class ForbiddenError(HTTPError):
    """Error code 403."""


class NotFoundError(HTTPError):
    """Error code 404."""


class MethodNotAllowedError(HTTPError):
    """Error code 405."""


class RateLimitError(HTTPError):
    """Error code 429."""


class GatewayError(HTTPError):
    """Error code 502."""


class ServerError(HTTPError):
    """
    Error code 5xx.
    Status code is not in the discord API
    """



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from enum import IntEnum
from typing import List, Optional, Tuple

from pincer.objects.user import User
from pincer.utils.api_object import APIObject
from pincer.utils.snowflake import Snowflake
from pincer.utils.types import MISSING, APINullable


class ActivityType(IntEnum):
    """
    :param GAME:
        Playing {name}
        e.g. "Playing Rocket League"

    :param STREAMING:
        Streaming {details}
        e.g. "Streaming Rocket League"
        Only supports Twitch and YouTube.

    :param LISTENING:
        Listening to {name}
        e.g. "Listening to Spotify"

    :param WATCHING:
        Watching {name}
        e.g. "Watching YouTube Together"

    :param CUSTOM:
        {emoji} {name}
        e.g. ":smiley: I am cool"

    :param COMPETING:
        Competing in {name}
        e.g. "Competing in Arena World Champions"
    """
    GAME = 0
    STREAMING = 1
    LISTENING = 2
    WATCHING = 3
    CUSTOM = 4
    COMPETING = 5


@dataclass
class ActivityTimestamp(APIObject):
    """
    :param start:
        unix time (in milliseconds) of when the activity started

    :param end:
        unix time (in milliseconds) of when the activity ends
    """
    start: APINullable[int] = MISSING
    end: APINullable[int] = MISSING


@dataclass
class ActivityEmoji(APIObject):
    """
    :param name:
        the name of the emoji

    :param id:
        the id of the emoji

    :param animated:
        whether this emoji is animated
    """
    name: str
    id: APINullable[Snowflake] = MISSING
    animated: APINullable[bool] = MISSING


@dataclass
class ActivityParty(APIObject):
    """
    :param id:
        the id of the party

    :param size:
        array of two integers (current_size, max_size)
    """
    id: APINullable[str] = MISSING
    size: APINullable[Tuple[int, int]] = MISSING


@dataclass
class ActivityAssets(APIObject):
    """
    :param large_image:
        the id for a large asset of the activity, usually a snowflake

    :param large_text:
        text displayed when hovering over
        the large image of the activity

    :param small_image:
        the id for a small asset of the activity, usually a snowflake

    :param small_text:
        text displayed when hovering over
        the small image of the activity
    """
    large_image: APINullable[str] = MISSING
    large_text: APINullable[str] = MISSING
    small_image: APINullable[str] = MISSING
    small_text: APINullable[str] = MISSING


@dataclass
class ActivitySecrets(APIObject):
    """
    :param join:
        the secret for joining a party

    :param spectate:
        the secret for spectating a game

    :param match:
        the secret for a specific instanced match
    """
    join: APINullable[str] = MISSING
    spectate: APINullable[str] = MISSING
    match_: APINullable[str] = MISSING


class ActivityFlags(IntEnum):
    INSTANCE = 1 << 0
    JOIN = 1 << 1
    SPECTATE = 1 << 2
    JOIN_REQUEST = 1 << 3
    SYNC = 1 << 4
    PLAY = 1 << 5


@dataclass
class ActivityButton(APIObject):
    """
    When received over the gateway, the buttons field is an array
    of strings, which are the button labels. Bots cannot access
    a user's activity button URLs. When sending, the buttons field
    must be an array of this object.

    :param label:
        the text shown on the button (1-32 characters)

    :param url:
        the url opened when clicking the button (1-512 characters)
    """
    label: str
    url: str


@dataclass
class Activity(APIObject):
    """
    Bots are only able to send `name`, `type`, and optionally `url`.

    :param name:
        the activity's name

    :param type:
        activity type

    :param url:
        stream url, is validated when type is 1

    :param created_at:
        unix timestamp (in milliseconds) of when
        the activity was added to the user's session

    :param timestamps:
        unix timestamps for start and/or end of the game

    :param application_id:
        application id for the game

    :param details:
        what the player is currently doing

    :param state:
        the user's current party status

    :param emoji:
        the emoji used for a custom status

    :param party:
        information for the current party of the player

    :param assets:
        images for the presence and their hover texts

    :param secrets:
        secrets for Rich Presence joining and spectating

    :param instance:
        whether or not the activity is an instanced game session

    :param flags:
        activity flags `OR`d together,
        describes what the payload includes
    """
    name: str
    type: ActivityType
    created_at: int

    url: APINullable[Optional[str]] = MISSING
    timestamps: APINullable[ActivityTimestamp] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    details: APINullable[Optional[str]] = MISSING
    state: APINullable[Optional[str]] = MISSING
    emoji: APINullable[Optional[ActivityEmoji]] = MISSING
    party: APINullable[ActivityParty] = MISSING
    assets: APINullable[ActivityAssets] = MISSING
    secrets: APINullable[ActivitySecrets] = MISSING
    instance: APINullable[bool] = MISSING
    flags: APINullable[ActivityFlags] = MISSING
    buttons: APINullable[List[ActivityButton]] = MISSING


@dataclass
class ClientStatus(APIObject):
    """
    Active sessions are indicated with an "online",
    "idle", or "dnd" string per platform.
    If a user is offline or invisible, the corresponding
    field is not present.

    :param desktop:
        the user's status set for an active desktop
        (Windows, Linux, Mac) application session

    :param mobile:
        the user's status set for an active mobile
        (iOS, Android) application session

    :param web:
        the user's status set for an active web
        (browser, bot account) application session
    """
    desktop: APINullable[str] = MISSING
    mobile: APINullable[str] = MISSING
    web: APINullable[str] = MISSING


@dataclass
class PresenceUpdateEvent(APIObject):
    """
    This event is sent when a user's presence or info,
    such as name or avatar, is updated.

    :param user:
        the user presence is being updated for

    :param guild_id:
        id of the guild

    :param status:
        either "idle", "dnd", "online", or "offline"

    :param activities:
        user's current activities

    :param client_status:
        user's platform-dependent status
    """
    user: User
    guild_id: Snowflake
    status: str
    activities: List[Activity]
    client_status: ClientStatus



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List

from .role import Role
from .user import User
from ..utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class Emoji(APIObject):
    """
    :param id:
        emoji id

    :param name:
        emoji name

    :param animated:
        whether this emoji is animated

    :param available:
        whether this emoji can be used, may be false due to loss of Server
        Boosts

    :param managed:
        whether this emoji is managed

    :param require_colons:
        whether this emoji must be wrapped in colons

    :param roles:
        roles allowed to use this emoji

    :param user:
        user that created this emoji
    """

    id: Optional[Snowflake]
    name: Optional[str]

    animated: APINullable[bool] = MISSING
    available: APINullable[bool] = MISSING
    managed: APINullable[bool] = MISSING
    require_colons: APINullable[bool] = MISSING
    roles: APINullable[List[Role]] = MISSING
    user: APINullable[User] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from typing import Optional

from .guild import Guild
from .user import User
from ..utils import APIObject, Snowflake, Timestamp


@dataclass
class GuildTemplate(APIObject):
    """
    Represents a code that when used,
    creates a guild based on a snapshot of an existing guild.

    :param code:
        the template code (unique ID)

    :param name:
        template name

    :param description:
        the description for the template

    :param usage_count:
        number of times this template has been used

    :param creator_id:
        the ID of the user who created the template

    :param creator: the
        user who created the template

    :param created_at:
        when this template was created

    :param updated_at:
        when this template was last synced to the source guild

    :param source_guild_id:
        the ID of the guild this template is based on

    :param serialized_source_guild:
        the guild snapshot this template contains

    :param is_dirty:
        whether the template has unsynced changes
    """
    code: str
    name: str
    description: Optional[str]
    usage_count: int
    creator_id: Snowflake
    creator: User
    created_at: Timestamp
    updated_at: Timestamp
    source_guild_id: Snowflake
    serialized_source_guild: Guild
    is_dirty: Optional[bool]



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from json import dumps, loads
from typing import Optional, Union, Dict, Any


class GatewayDispatch:
    """Represents a websocket message."""

    def __init__(
            self,
            op: int,
            data: Optional[Union[int, Dict[str, Any]]],
            seq: Optional[int] = None,
            name: Optional[str] = None
    ):
        """
        Instantiate a new GatewayDispatch object.

        :param op:
            The discord opcode which represents what the message means.

        :param data:
            The event data that has been sent/received.

        :param seq:
            The sequence number of a message, which can be used
            for resuming sessions and heartbeats.

        :param name:
            The event name for the payload.
        """
        self.op: int = op
        self.data: Optional[Union[int, Dict[str, Any]]] = data
        self.seq: Optional[int] = seq
        self.event_name: Optional[str] = name

    def __str__(self) -> str:
        """
        :return
            The string representation of the GatewayDispatch object.

        This can be used to send a websocket message to the gateway.
        """
        return dumps(
            dict(
                op=self.op,
                d=self.data,
                s=self.seq,
                t=self.event_name
            )
        )

    @classmethod
    def from_string(cls, payload: str) -> GatewayDispatch:
        """
        Parses a given payload from a string format
            and returns a GatewayDispatch.

        :param payload:
            The payload to parse.

        :return:
            A proper GatewayDispatch object.
        """
        payload: Dict[str, Union[int, str, Dict[str, Any]]] = loads(payload)
        return cls(
            payload.get("op"),
            payload.get("d"),
            payload.get("s"),
            payload.get("t")
        )



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import logging
from asyncio import sleep
from typing import Optional

from websockets.legacy.client import WebSocketClientProtocol

from . import __package__
from ..core.dispatch import GatewayDispatch
from ..exceptions import HeartbeatError

_log = logging.getLogger(__package__)


class Heartbeat:
    """
    The heartbeat of the websocket connection.

    This is what lets the server and client know that they are still
        both online and properly connected.
    """
    __heartbeat: float = 0
    __sequence: Optional[int] = None

    @classmethod
    async def __send(cls, socket: WebSocketClientProtocol):
        """
        Sends a heartbeat to the API gateway.
        :meta public:

        :param socket:
            The socket to send the heartbeat to.
        """
        _log.debug("Sending heartbeat (seq: %s)", str(cls.__sequence))
        await socket.send(str(GatewayDispatch(1, cls.__sequence)))

    @classmethod
    def get(cls) -> float:
        """
        Get the current heartbeat.

        :return:
            The current heartbeat of the client.
            Default is 0 (client has not initialized the heartbeat yet.)
        """
        return cls.__heartbeat

    @classmethod
    async def handle_hello(
            cls,
            socket: WebSocketClientProtocol,
            payload: GatewayDispatch
    ):
        """
        Handshake between the discord API and the client.
        Retrieve the heartbeat for maintaining a connection.

        :param socket:
            The socket to send the heartbeat to.

        :param payload:
            The received hello message from the Discord gateway.
        """
        _log.debug("Handling initial discord hello websocket message.")
        cls.__heartbeat = payload.data.get("heartbeat_interval")

        if not cls.__heartbeat:
            _log.error(
                "No `heartbeat_interval` is present. Has the API changed? "
                "(payload: %s)", payload
            )

            raise HeartbeatError(
                "Discord hello is missing `heartbeat_interval` in payload."
                "Because of this the client can not maintain a connection. "
                "Check logging for more information."
            )

        cls.__heartbeat /= 1000

        _log.debug(
            "Maintaining a connection with heartbeat: %s", cls.__heartbeat
        )

        if Heartbeat.__sequence:
            await socket.send(
                str(GatewayDispatch(6, cls.__sequence, seq=cls.__sequence))
            )

        else:
            await cls.__send(socket)

    @classmethod
    async def handle_heartbeat(cls, socket: WebSocketClientProtocol, _):
        """
        Handles a heartbeat, which means that it rests
        and then sends a new heartbeat.

        :param socket:
            The socket to send the heartbeat to.

        :param _:
            Filling param for auto event handling.
        """

        _log.debug("Resting heart for %is", cls.__heartbeat)
        await sleep(cls.__heartbeat)
        await cls.__send(socket)

    @classmethod
    def update_sequence(cls, seq: int):
        """
        Update the heartbeat sequence.

        :param seq:
            The new heartbeat sequence to be updated with.
        """
        _log.debug("Updating heartbeat sequence...")
        cls.__sequence = seq



# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from re import match
from typing import Any, Callable, Dict, Iterable, Union, Optional

from ..exceptions import InvalidUrlError, EmbedFieldError
from ..utils import APIObject, APINullable, MISSING


def _field_size(_field: str) -> int:
    """
    The Discord API removes white space
        when counting the length of a field.

    :param _field:
        The field.

    :return:
        Length of the string without white space.
    """
    return 0 if _field == MISSING else len(_field.strip())


def _is_valid_url(url: str) -> bool:
    """
    Checks whether the url is a proper and valid url.
    (matches for http and attachment protocol.

    :param url:
        The url which must be checked.

    :return:
        Whether the provided url is valid.
    """
    stmt = (
        r"(http[s]|attachment)"
        r"?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|"
        r"(?:%[0-9a-fA-F][0-9a-fA-F]))+"
    )

    return bool(match(stmt, url))


def _check_if_valid_url(url: str):
    """
    Checks if the provided url is valid.

    :raises InvalidUrlError:
        if the url didn't match the url regex.
        (which means that it was malformed or didn't match the http/attachment
        protocol).
    """
    if not _is_valid_url(url):
        raise InvalidUrlError(
            "Url was malformed or wasn't of protocol http(s)/attachment."
        )


@dataclass
class EmbedAuthor:
    """
    Representation of the Embed Author class

    :param name:
        Name of the author

    :param url:
        Url of the author

    :param icon_url:
        Url of the author icon

    :param proxy_icon_url:
        A proxied url of the author icon
    """
    icon_url: APINullable[str] = MISSING
    name: APINullable[str] = MISSING
    proxy_icon_url: APINullable[str] = MISSING
    url: APINullable[str] = MISSING

    def __post_init__(self):
        """
        :raises EmbedFieldError:
            Name is longer than 256 characters.

        :raises InvalidUrlError:
            if the url didn't match the url regex.
            (which means that it was malformed or didn't match the
            http/attachment protocol).
        """
        if _field_size(self.name) > 256:
            raise EmbedFieldError.from_desc("Author name", 256, len(self.name))

        _check_if_valid_url(self.url)


@dataclass
class EmbedImage:
    """
    Representation of the Embed Image class

    :param url:
        Source url of the image

    :param proxy_url:
        A proxied url of the image

    :param height:
        Height of the image

    :param width:
        Width of the image
    """

    height: APINullable[int] = MISSING
    proxy_url: APINullable[str] = MISSING
    url: APINullable[str] = MISSING
    width: APINullable[int] = MISSING

    def __post_init__(self):
        """
        :raises InvalidUrlError:
            if the url didn't match the url regex.
            (which means that it was malformed or didn't match the
            http/attachment protocol).
        """
        _check_if_valid_url(self.url)


@dataclass
class EmbedProvider:
    """
    Representation of the Provider class

    :param name:
        Name of the provider

    :param url:
        Url of the provider
    """
    name: APINullable[str] = MISSING
    url: APINullable[str] = MISSING


@dataclass
class EmbedThumbnail:
    """
    Representation of the Embed Thumbnail class

    :param url:
        Source url of the thumbnail

    :param proxy_url:
        A proxied url of the thumbnail

    :param height:
        Height of the thumbnail

    :param width:
        Width of the thumbnail

    """

    height: APINullable[int] = MISSING
    proxy_url: APINullable[str] = MISSING
    url: APINullable[str] = MISSING
    width: APINullable[int] = MISSING

    def __post_init__(self):
        """
        :raises InvalidUrlError:
            if the url didn't match the url regex.
            (which means that it was malformed or didn't match the
            http/attachment protocol).
        """
        _check_if_valid_url(self.url)


@dataclass
class EmbedVideo:
    """
    Representation of the Embed Video class

    :param url:
        Source url of the video

    :param proxy_url:
        A proxied url of the video

    :param height:
        Height of the video

    :param width:
        Width of the video
    """
    height: APINullable[int] = MISSING
    url: APINullable[str] = MISSING
    proxy_url: APINullable[str] = MISSING
    width: APINullable[int] = MISSING


@dataclass
class EmbedFooter:
    """
    Representation of the Embed Footer class

    :param text:
        Footer text

    :param icon_url:
        Url of the footer icon

    :param proxy_icon_url:
        A proxied url of the footer icon

    :raises EmbedFieldError:
        Text is longer than 2048 characters
    """

    text: str

    icon_url: APINullable[str] = MISSING
    proxy_icon_url: APINullable[str] = MISSING

    def __post_init__(self):
        if _field_size(self.text) > 2048:
            raise EmbedFieldError.from_desc(
                "Footer text", 2048, len(self.text)
            )


@dataclass
class EmbedField:
    """
    Representation of the Embed Field class

    :param name:
        The name of the field

    :param value:
        The text in the field

    :param inline:
        Whether or not this field should display inline

    :raises EmbedFieldError:
        Name is longer than 256 characters

    :raises EmbedFieldError:
        Description is longer than 1024 characters
    """

    name: str
    value: str

    inline: APINullable[bool] = MISSING

    def __post_init__(self):
        if _field_size(self.name) > 256:
            raise EmbedFieldError.from_desc(
                "Field name", 256, len(self.name)
            )

        if _field_size(self.value) > 1024:
            raise EmbedFieldError.from_desc(
                "Field value", 1024, len(self.value)
            )


# TODO: Handle Bad Request if embed that is too big is sent
# https://discord.com/developers/docs/resources/channel#embed-limits
# Currently ignored since I don't think it would make sense to put
# This with the Embed class
@dataclass
class Embed(APIObject):
    """
    Representation of the discord Embed class

    :property author:
        Author information.

    :param color:
        Embed color code.

    :param description:
        Embed description.

    :param fields:
        Fields information.

    :param footer:
        Footer information.

    :param image:
        Image information.

    :param provider:
        Provider information.

    :param thumbnail:
        Thumbnail information.

    :param timestamp:
        Timestamp of embed content in ISO format.

    :param title:
        Embed title.

    :param url:
        Embed url.

    :param video:
        Video information.
    """

    author: APINullable[EmbedAuthor] = MISSING
    color: APINullable[int] = MISSING
    description: APINullable[str] = MISSING
    fields: list[EmbedField] = field(default_factory=list)
    footer: APINullable[EmbedFooter] = MISSING
    image: APINullable[EmbedImage] = MISSING
    provider: APINullable[EmbedProvider] = MISSING
    thumbnail: APINullable[EmbedThumbnail] = MISSING
    timestamp: APINullable[str] = MISSING
    title: APINullable[str] = MISSING
    url: APINullable[str] = MISSING
    video: APINullable[EmbedVideo] = MISSING

    def __post_init__(self):
        """
        :raises EmbedFieldError:
            Embed title is longer than 256 characters.
        :raises EmbedFieldError:
            Embed description is longer than 4096 characters.
        :raises EmbedFieldError:
            Embed has more than 25 fields.
        """
        if _field_size(self.title) > 256:
            raise EmbedFieldError.from_desc(
                "Embed title", 256, len(self.title)
            )

        if _field_size(self.description) > 4096:
            raise EmbedFieldError.from_desc(
                "Embed description", 4096, len(self.description)
            )

        if len(self.fields) > 25:
            raise EmbedFieldError.from_desc(
                "Embed field", 25, len(self.fields)
            )

    def set_timestamp(self, time: datetime) -> Embed:
        """
        Discord uses iso format for time stamps.
        This function will set the time to that format.

        :param time:
            A datetime object.

        :return: self
        """
        self.timestamp = time.isoformat()

        return self

    def set_author(
            self,
            icon_url: APINullable[str] = MISSING,
            name: APINullable[str] = MISSING,
            proxy_icon_url: APINullable[str] = MISSING,
            url: APINullable[str] = MISSING
    ) -> Embed:
        """
        Set the author message for the embed. This is the top
        field of the embed.

        :param icon_url:
            The icon which will be next to the author name.

        :param name:
            The name for the author (so the message).

        :param proxy_icon_url:
            A proxied url of the author icon.

        :param url:
            The url for the author name, this will make the
            name field a link/url.

        :return: self
        """

        self.author = EmbedAuthor(
            icon_url=icon_url,
            name=name,
            proxy_icon_url=proxy_icon_url,
            url=url
        )

        return self

    def set_image(
            self,
            height: APINullable[int] = MISSING,
            url: APINullable[str] = MISSING,
            proxy_url: APINullable[str] = MISSING,
            width: APINullable[int] = MISSING
    ) -> Embed:
        """
        Sets an image for your embed.

        :param url:
            Source url of the video

        :param proxy_url:
            A proxied url of the video

        :param height:
            Height of the video

        :param width:
            Width of the video

        :return: self
        """
        self.image = EmbedImage(
            height=height,
            url=url,
            proxy_url=proxy_url,
            width=width
        )

        return self

    def set_thumbnail(
            self,
            height: APINullable[int] = MISSING,
            url: APINullable[str] = MISSING,
            proxy_url: APINullable[str] = MISSING,
            width: APINullable[int] = MISSING
    ) -> Embed:
        """
        Sets the thumbnail of the embed.
        This image is bigger than the `image` property.

        :param url:
            Source url of the video

        :param proxy_url:
            A proxied url of the video

        :param height:
            Height of the video

        :param width:
            Width of the video

        :return self:
        """
        self.thumbnail = EmbedThumbnail(
            height=height,
            url=url,
            proxy_url=proxy_url,
            width=width
        )

        return self

    def set_footer(
            self,
            text: str,
            icon_url: APINullable[str] = MISSING,
            proxy_icon_url: APINullable[str] = MISSING
    ) -> Embed:
        """
        Sets the embed footer. This is at the bottom of your embed.

        :param text:
            Footer text

        :param icon_url:
            Url of the footer icon

        :param proxy_icon_url:
            A proxied url of the footer icon

        :return: self
        """
        self.footer = EmbedFooter(
            text=text,
            icon_url=icon_url,
            proxy_icon_url=proxy_icon_url
        )

        return self

    def add_field(
            self,
            name: str,
            value: str,
            inline: APINullable[bool] = MISSING
    ) -> Embed:
        """
        Adds a field to the embed.
        An embed can contain up to 25 fields.

        :param name:
            The name of the field

        :param value:
            The text in the field

        :param inline:
            Whether or not this field should display inline

        :raises EmbedFieldError:
            Raised when there are more than 25 fields in the embed
        """
        _field = EmbedField(
            name=name,
            value=value,
            inline=inline
        )

        if len(self.fields) > 25:
            raise EmbedFieldError.from_desc(
                "Embed field", 25, len(self.fields) + 1
            )

        self.fields += [_field]

        return self

    def add_fields(
        self,
        field_list: Union[Dict[Any, Any], Iterable[Iterable[Any, Any]]],
        checks: Optional[Callable[[Any], Any]] = bool,
        map_title: Optional[Callable[[Any], str]] = str,
        map_values: Optional[Callable[[Any], str]] = str,
        inline: bool = True
    ) -> Embed:
        """
        Add multiple fields from a list,
        dict or generator of fields with possible mapping.

        :param field_list:
            A iterable or generator of the fields to add.
            If the field_list type is a dictionary, will take items.

        :param checks:
            A filter function to remove embed fields.

        :param map_title:
            A transform function to change the titles.

        :param map_values:
            A transform function to change the values.

        :param inline:
            Whether to create grid or each field on a new line.

        :raises EmbedFieldError:
            Raised when there are more than 25 fields in the embed

        :return: the embed for chaining methods.
        """

        if isinstance(field_list, dict):
            field_list: Iterable[Iterable[Any, Any]] = field_list.items()

        for field_name, field_value in field_list:
            val = (
                map_values(field_value)
                if not isinstance(field_value, tuple)
                else map_values(*field_value)
            )

            if checks(val):
                self.add_field(
                    name=map_title(field_name),
                    value=val,
                    inline=inline
                )

        return self



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from dataclasses import dataclass

from ..utils import APIObject, APINullable, MISSING, Snowflake


@dataclass
class RoleTags(APIObject):
    """
    Special tags/flags which have been defined for a role.

    :bot_id:
        The id of the bot this role belongs to.
        (the role got created by adding the bot with this id)

    :integration_id:
        The id of the integration this role belongs to.
        (the role got created by adding an integration with this id)

    :premium_subscriber:
        Whether this is the guild's premium subscriber role or not.
    """
    bot_id: APINullable[Snowflake] = MISSING
    integration_id: APINullable[Snowflake] = MISSING
    premium_subscriber: APINullable[bool] = MISSING


@dataclass
class Role(APIObject):
    """
    Represents a Discord guild/server role.

    :param color:
        integer representation of hexadecimal color code

    :param hoist:
        if this role is pinned in the user listing

    :param id:
        role id

    :param managed:
        whether this role is managed by an integration

    :param mentionable:
        whether this role is mentionable

    :param name:
        role name

    :param permissions:
        permission bit set

    :param position:
        position of this role

    :param tags:
        the tags this role has
    """

    color: int
    hoist: bool
    id: Snowflake
    managed: bool
    mentionable: bool
    name: str
    permissions: str
    position: int

    tags: APINullable[RoleTags] = MISSING



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2021 Pincer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from dataclasses import dataclass
from enum import IntEnum
from typing import Dict

from .user import User
from ..utils import APIObject, Snowflake, APINullable, MISSING


class CallbackType(IntEnum):
    """
    The types of response a client can give to a interaction.

    :param PONG:
        ACK a Ping

    :param MESSAGE:
        respond to an interaction with a message

    :param DEFERRED_MESSAGE:
        ACK an interaction and edit a response later,
        the user sees a loading state

    :param DEFERRED_UPDATE_MESSAGE:
        for components, ACK an interaction
        and edit the original message later;
        the user does not see a loading state

    :param UPDATE_MESSAGE
        for components, edit the message the component was attached to
    """
    PONG = 1
    MESSAGE = 4
    DEFERRED_MESSAGE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7


class InteractionType(IntEnum):
    """
    Represents the different types of interactions the client
    can have with a member.

    :param CHAT_INPUT:
        Slash commands;
        a text-based command that shows up when a user types /

    :param USER:
        A UI-based command that shows up when you right click
        or tap on a user

    :param MESSAGE:
        A UI-based command that shows up when you right click
        or tap on a message
    """
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3


@dataclass
class MessageInteraction(APIObject):
    """
    Represents a Discord Message Interaction object

    This is sent on the message object when the message
    is a response to an Interaction without an existing message.

    :param id: id of the interaction
    :param type: the type of interaction
    :param name: the name of the application command
    :param user: the user who invoked the interaction
    """
    id: Snowflake
    type: InteractionType
    name: str
    user: User

    member: APINullable[Dict] = MISSING
